# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
"""


import numpy as np
from numpy.testing import assert_equal

import unittest

from chemometrics import mcr
from chemometrics.mcr import McrAR
import chemometrics.mcr.constraint as constraint


def dataset():
    """ Setups dataset """

    M = 21
    N = 21
    P = 101
    n_components = 2

    C_img = np.zeros((M, N, n_components))
    C_img[..., 0] = np.dot(np.ones((M, 1)), np.linspace(0, 1, N)[None, :])
    C_img[..., 1] = 1 - C_img[..., 0]

    St_known = np.zeros((n_components, P))
    St_known[0, 40:60] = 1
    St_known[1, 60:80] = 2

    C_known = C_img.reshape((-1, n_components))

    D_known = np.dot(C_known, St_known)

    return C_known, D_known, St_known


class TestMcrSklearnLike(unittest.TestCase):
    """
    Test sklearn-like behavior of mcr module.

    This test series ensures that a kwargs initialisation is possible of McrAR.
    It seems to more or less be a copy of test_mcr but now with kwargs.

    Potentially integrate into test_mcr in future.
    """

    def test_sklearn_mcr_ideal_default(self):
        """ Provides C/St_known so optimal should be 1 iteration """

        C_known, D_known, St_known = dataset()

        mcrar = McrAR(fit_kwargs={'ST': St_known})
        mcrar.fit(D_known)
        self.assertLess(mcrar.n_iter_opt, 3)
        self.assertTrue(((mcrar.D_ - D_known)**2).mean() < 1e-10)
        self.assertTrue(((mcrar.D_opt_ - D_known)**2).mean() < 1e-10)

        mcrar = McrAR(fit_kwargs={'C': C_known})
        mcrar.fit(D_known)
        self.assertLess(mcrar.n_iter_opt, 3)
        self.assertTrue(((mcrar.D_ - D_known)**2).mean() < 1e-10)
        self.assertTrue(((mcrar.D_opt_ - D_known)**2).mean() < 1e-10)

    def test_sklearn_mcr_ideal_str_regressors(self):
        """ Test MCR with string-provded regressors"""

        C_known, D_known, St_known = dataset()

        mcrar = McrAR(c_regr='OLS', st_regr='OLS', fit_kwargs={'ST': St_known})
        mcrar.fit(D_known, verbose=True)
        assert_equal(1, mcrar.n_iter_opt)
        self.assertIsInstance(mcrar.c_regressor, mcr.regressor.OLS)
        self.assertIsInstance(mcrar.st_regressor, mcr.regressor.OLS)

        mcrar = McrAR(c_regr='NNLS', st_regr='NNLS',
                      fit_kwargs={'ST': St_known})
        mcrar.fit(D_known)
        assert_equal(1, mcrar.n_iter_opt)
        self.assertIsInstance(mcrar.c_regressor, mcr.regressor.NNLS)
        self.assertIsInstance(mcrar.st_regressor, mcr.regressor.NNLS)
        self.assertTrue(((mcrar.D_ - D_known)**2).mean() < 1e-10)
        self.assertTrue(((mcrar.D_opt_ - D_known)**2).mean() < 1e-10)

        # Provided C_known this time
        mcrar = McrAR(c_regr='OLS', st_regr='OLS', fit_kwargs={'C': C_known})
        mcrar.fit(D_known)

        # Turns out some systems get it in 1 iteration, some in 2
        # assert_equal(1, mcrar.n_iter_opt)
        assert_equal(True, mcrar.n_iter_opt <= 2)

        self.assertTrue(((mcrar.D_ - D_known)**2).mean() < 1e-10)
        self.assertTrue(((mcrar.D_opt_ - D_known)**2).mean() < 1e-10)

    def test_sklearn_mcr_max_iterations(self):
        """ Test MCR exits at max_iter"""

        C_known, D_known, St_known = dataset()

        # Seeding with a constant of 0.1 for C, actually leads to a bad local
        # minimum; thus, the err_change gets really small with a relatively bad
        # error. The tol_err_change is set to None, so it makes it to max_iter.
        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=None, tol_n_increase=None,
                      tol_err_change=None, tol_n_above_min=None,
                      fit_kwargs={'C': C_known*0 + 0.1})
        mcrar.fit(D_known)
        self.assertTrue(mcrar.exit_max_iter_reached)

    def test_sklearn_mcr_tol_increase(self):
        """ Test MCR exits due error increasing above a tolerance fraction"""

        C_known, D_known, St_known = dataset()

        # Seeding with a constant of 0.1 for C, actually leads to a bad local
        # minimum; thus, the err_change gets really small with a relatively bad
        # error.
        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=0, tol_n_increase=None,
                      tol_err_change=None, tol_n_above_min=None,
                      fit_kwargs={'C': C_known*0 + 0.1})
        mcrar.fit(D_known)
        self.assertTrue(mcrar.exit_tol_increase)

    def test_sklearn_mcr_tol_n_increase(self):
        """
        Test MCR exits due iterating n times with an increase in error

        Note: On some CI systems, the minimum err bottoms out; thus,
        tol_n_above_min needed to be set to 0 to trigger a break.
        """

        C_known, D_known, St_known = dataset()

        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=None, tol_n_increase=0,
                      tol_err_change=None, tol_n_above_min=None,
                      fit_kwargs={'C': C_known*0 + 0.1})
        mcrar.fit(D_known)
        self.assertTrue(mcrar.exit_tol_n_increase)

    def test_sklearn_mcr_tol_err_change(self):
        """ Test MCR exits due error increasing by a value """

        C_known, D_known, St_known = dataset()

        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=None, tol_n_increase=None,
                      tol_err_change=1e-20, tol_n_above_min=None,
                      fit_kwargs={'C': C_known})
        mcrar.fit(D_known)
        self.assertTrue(mcrar.exit_tol_err_change)

    def test_sklearn_mcr_tol_n_above_min(self):
        """
        Test MCR exits due to half-terating n times with error above the
        minimum error.

        Note: On some CI systems, the minimum err bottoms out; thus,
        tol_n_above_min needed to be set to 0 to trigger a break.
        """

        C_known, D_known, St_known = dataset()

        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=None, tol_n_increase=None,
                      tol_err_change=None, tol_n_above_min=0,
                      fit_kwargs={'C': C_known*0 + 0.1})
        mcrar.fit(D_known)
        self.assertTrue(mcrar.exit_tol_n_above_min)

    def test_sklearn_mcr_st_semilearned(self):
        """ Test when St items are fixed, i.e., enforced to be the same as the
        input, always """

        M = 21
        N = 21
        P = 101
        n_components = 3

        C_img = np.zeros((M, N, n_components))
        C_img[..., 0] = np.dot(np.ones((M, 1)), np.linspace(0, 1, N)[None, :])
        C_img[..., 1] = np.dot(np.linspace(0, 1, M)[:, None], np.ones((1, N)))
        C_img[..., 2] = 1 - C_img[..., 0] - C_img[..., 1]
        C_img = C_img / C_img.sum(axis=-1)[:, :, None]

        St_known = np.zeros((n_components, P))
        St_known[0, 30:50] = 1
        St_known[1, 50:70] = 2
        St_known[2, 70:90] = 3
        St_known += 1

        C_known = C_img.reshape((-1, n_components))

        D_known = np.dot(C_known, St_known)

        ST_guess = 1 * St_known
        ST_guess[2, :] = np.random.randn(P)

        mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_err_change=1e-10,
                      fit_kwargs={'ST': ST_guess, 'st_fix': [0, 1]})

        mcrar.fit(D_known)
        assert_equal(mcrar.ST_[0, :], St_known[0, :])
        assert_equal(mcrar.ST_[1, :], St_known[1, :])

    def test_sklearn_mcr_c_semilearned(self):
        """ Test when C items are fixed, i.e., enforced to be the same as the
        input, always """

        M = 21
        N = 21
        P = 101
        n_components = 3

        C_img = np.zeros((M, N, n_components))
        C_img[..., 0] = np.dot(np.ones((M, 1)), np.linspace(0, 1, N)[None, :])
        C_img[..., 1] = np.dot(np.linspace(0, 1, M)[:, None], np.ones((1, N)))
        C_img[..., 2] = 1 - C_img[..., 0] - C_img[..., 1]
        C_img = C_img / C_img.sum(axis=-1)[:, :, None]

        St_known = np.zeros((n_components, P))
        St_known[0, 30:50] = 1
        St_known[1, 50:70] = 2
        St_known[2, 70:90] = 3
        St_known += 1

        C_known = C_img.reshape((-1, n_components))

        D_known = np.dot(C_known, St_known)

        C_guess = 1 * C_known
        C_guess[:, 2] = np.abs(np.random.randn(int(M*N))+0.1)

        mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_err_change=1e-10,
                      fit_kwargs={'C': C_guess, 'c_fix': [0, 1]})

        mcrar.fit(D_known)
        assert_equal(mcrar.C_[:, 0], C_known[:, 0])
        assert_equal(mcrar.C_[:, 1], C_known[:, 1])

    def test_sklearn_mcr_semilearned_both_c_st(self):
        """
        Test the special case when C & ST are provided, requiring C-fix ST-fix
        to be provided
        """

        M = 21
        N = 21
        P = 101
        n_components = 3

        C_img = np.zeros((M, N, n_components))
        C_img[..., 0] = np.dot(np.ones((M, 1)),
                               np.linspace(0.1, 1, N)[None, :])
        C_img[..., 1] = np.dot(np.linspace(0.1, 1, M)[:, None],
                               np.ones((1, N)))
        C_img[..., 2] = 1 - C_img[..., 0] - C_img[..., 1]
        C_img = C_img / C_img.sum(axis=-1)[:, :, None]

        St_known = np.zeros((n_components, P))
        St_known[0, 30:50] = 1
        St_known[1, 50:70] = 2
        St_known[2, 70:90] = 3
        St_known += 1

        C_known = C_img.reshape((-1, n_components))

        D_known = np.dot(C_known, St_known)

        C_guess = 1 * C_known
        C_guess[:, 2] = np.abs(np.random.randn(int(M*N)))

        mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_err_change=1e-10,
                      fit_kwargs={'C': C_guess,
                                  'ST': St_known,
                                  'c_fix': [0, 1],
                                  'st_fix': [0]})

        mcrar.fit(D_known, c_first=True)
        assert_equal(mcrar.C_[:, 0], C_known[:, 0])
        assert_equal(mcrar.C_[:, 1], C_known[:, 1])
        assert_equal(mcrar.ST_[0, :], St_known[0, :])

        # ST-solve first
        mcrar.fit(D_known, C=C_guess, ST=St_known, c_fix=[0, 1], st_fix=[0],
                  c_first=False)
        assert_equal(mcrar.C_[:, 0], C_known[:, 0])
        assert_equal(mcrar.C_[:, 1], C_known[:, 1])
        assert_equal(mcrar.ST_[0, :], St_known[0, :])

    def test_sklearn_mcr_errors(self):

        # Providing C in fit_kwargs and S^T to fit without both C_fix and
        # St_fix
        with self.assertRaises(TypeError):
            mcrar = McrAR(fit_kwargs={'C': np.random.randn(10, 3),
                                      'c_fix': [0]})
            # Only c_fix
            mcrar.fit(np.random.randn(10, 5), ST=np.random.randn(3, 5))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.


import matplotlib.pyplot as plt
import matplotlib.cm
import numpy as np
from sklearn.utils.extmath import randomized_svd


def plot_colored_series(Y, x=None, reference=None):
    r"""
    Plot lines colored by position or `reference`

    Generate a line plot with `x` on x-axis and one or multiple dataseries
    `Y`. The lines are either colored by position in the matrix `Y` or by
    value in the `reference` matrix.

    Parameters
    ----------
    Y : (n, m) ndarray
        Matrix containing data series to plot. The function expects. `n`
        datapoints in `m` series.
    x : {None, (n,) ndarray}
        Location on x-axis
    reference : {None (default), (m,) ndarray}
        Reference values to color data series by. If None, the series are
        colored by the position in the second dimension of matrix `Y`.

    Returns
    -------
    lines : list
        A list of line objects generated by plotting the spectra.
    """
    # define number of input series for line plot
    if (Y.ndim > 1):
        n_series = Y.shape[1]
    else:
        n_series = 1
    if x is None:
        x = np.arange(Y.shape[0])
    # if no reference is given a dummy reference is needed (sequential
    # coloring)
    if reference is None:
        reference = np.arange(n_series)
    myMapper = matplotlib.cm.ScalarMappable(cmap='viridis')
    colors = myMapper.to_rgba(reference)
    lines = []
    for i in range(n_series):
        line_i = plt.plot(x, Y[:, i], color=colors[i, :])
        lines.append(line_i[0])
    return lines


def plot_svd(D, n_comp=5, n_eigenvalues=20):
    r"""
    Plot SVD-matrices in three subplots.

    Perform a Singular Value Decomposition (SVD) and plot the three matrices
    in three subplots. The number of singular vectors shown is ``n_comp``. The
    left subplot contains the left singular vectors, the middle subplot the
    singular values, the right subplot the right singular vectors. The
    function is a useful tool to get first insights into a data set. It helps
    to evaluate which components contain information and which mainly noise.
    Compared to Principal Component Analysis (PCA), the singular vectors are
    normalized and scaling results from the eigenvalues.

    Parameters
    ----------
    D : (n, m) ndarray
        Matrix containing data to plot and analyze. The function expects. ``n``
        samples with ``m`` signals (e.g. wavelengths, measurements).

    n_comp : int
        Number of singular vectors to plot.

    n_eigenvalues : int

    Returns
    -------
    fig : figure
        A list of line objects generated by plotting the spectra.

    References
    ----------
    Plot style adapted from personal communication with Matthias Sawall as
    in Figure 5 [1]_.

    .. [1] M. Sawall, A. Börner, C. Kubis, D. Selent, R. Ludwig, and K.
           Neymeyr. Model-free multivariate curve resolution combined with
           model-based kinetics: Algorithm and applications. J. Chemom.,
           26:538–548, 2012.
    """
    u, s, vh = randomized_svd(D, n_components=n_eigenvalues, random_state=None)

    _ = plt.figure(figsize=(15, 5))
    plt.subplot(131)
    plt.plot(u[:, :n_comp])
    plt.subplot(132)
    for i in range(n_eigenvalues):
        if i < n_comp:
            plt.plot(i, s[i], 'o')
        else:
            plt.plot(i, s[i], 'ok')
    plt.gca().set_yscale('log')
    plt.subplot(133)
    plt.plot(vh.T[:, :n_comp])



# Copyright 2021, 2022 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import chemometrics as cm  # accounts for relativ path
import numpy as np
import unittest
from numpy.testing import assert_allclose
from scipy.interpolate import interp1d


class TestGenerate_spectra(unittest.TestCase):
    r"""
    Test the `generate_spectra` function.
    """

    def test_shape(self):
        """
        Test if correct shape is generated
        """
        n_wl = 200
        expected_shape = (n_wl,)
        output_shape = cm.generate_spectra(n_wl, 2, 50).shape
        self.assertTrue(np.all(expected_shape == output_shape))

    def test_no_bands(self):
        """
        Test if ``n_band = 0`` returns zero vector.
        """
        n_wl = 10
        n_bands = 0
        bandwidth = 100
        spectra = cm.generate_spectra(n_wl, n_bands, bandwidth)
        isZero = np.all(np.isclose(np.zeros(n_wl), spectra))
        self.assertTrue(isZero)


class TestPsaudoVoigt(unittest.TestCase):
    r"""
    Test Pseudo Voigt profiles
    """

    def test_shape(self):
        """
        Test if _pseudo_voigt returns correct shape.
        """

        wl = np.arange(1000)[:, None]
        parameters = np.array([
            [200, 0.5, 50, 500],
            [100, 0.8, 20, 200]
        ]).T

        spec = cm.pseudo_voigt_spectra(wl, parameters)

        self.assertTrue(spec.shape == wl.shape)

    def test_statistics_pdf(self):
        """
        Test area, mode and full width at half max of a single Voigt peak
        in lorentzian limit
        """
        wl = np.arange(1000)[:, None]
        par_list = [
                np.array([[500, 1, 0.5, 10], ]).T,
                np.array([[500, 1, 0.5, 20], ]).T,
                np.array([[280, 1, 0.2, 30], ]).T
        ]

        for parameters in par_list:
            spec = cm.pseudo_voigt_spectra(wl, parameters)

            # calculate mode (wl_max)
            wl_max = wl[np.argmax(spec)][0]
            assert_allclose(parameters[0], wl_max)
            signal_max = np.max(spec)
            assert_allclose(parameters[1], signal_max)

            # calculate fwhm by inverting half (>wl_max) of the
            # propability  density function and analyzing width

            invPDF = interp1d(spec[wl_max:, 0], wl[wl_max:, 0])
            fwhm = 2*(invPDF(signal_max/2)-wl_max)

            assert_allclose(fwhm, 2*parameters[3])



# Copyright 2022 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import matplotlib
from scipy.stats import binom
import numpy as np


class TestLVmixin():
    """
    A mixin for running LVmixin tests
    """
    def test_dmodx_shape(self):
        """
        Test that dmodx provides a vector of the correct shape
        """
        dmodx = self.model.dmodx(self.X)
        self.assertTrue(dmodx.shape == (self.n_samples, ))

        dmodx = self.model.dmodx(self.X, normalize=False)
        self.assertTrue(dmodx.shape == (self.n_samples, ))

        dmodx = self.model.dmodx(self.X, absolute=True)
        self.assertTrue(dmodx.shape == (self.n_samples, ))

    def test_dmodx_length(self):
        """
        Test that dmodx = 0 if data from hyperplan is taken
        """
        X_hat = self.model.inverse_transform(self.model.transform(self.X))
        dmodx = self.model.dmodx(X_hat, normalize=False)
        self.assertTrue(np.allclose(dmodx, 0))

    def test_score_plus_dmodx_length(self):
        """
        Test that the variation on the model plane and the variation
        orthogonal to the model plane yield the total variation.
        """
        dmodx = self.model.dmodx(self.X, normalize=False)
        absolut_dmodx = dmodx**2 * (self.X.shape[1] - self.model.n_components)
        X_hat_bar = self.model.x_scores_ @ self.model.x_loadings_.T
        ss_X_hat = np.sum(X_hat_bar**2, axis=1)
        X_bar = self.X - np.mean(self.X, axis=0)
        ss_X = np.sum(X_bar**2, axis=1)
        self.assertTrue(np.allclose(ss_X_hat + absolut_dmodx, ss_X,
                                    atol=1e-2))

    def test_distance_plot_return_arg(self):
        """
        Test that distance_plot returns an axis cv_object
        """
        ax = self.model.distance_plot(self.X)
        self.assertIsInstance(ax, list)
        for inst in ax:
            self.assertIsInstance(inst, matplotlib.axes.Subplot)

    def test_crit_dmodx(self):
        """
        Test that number of outliers corresponds

        Performs a binomial statistical test. This test fails more often then
        the confidence level would predict.

        Potential explanation:
        Since dmodx
        is only approximately f2 distributed [1], crit_dmodx is not completely
        accurate. Especially when considering outliers at high confidence, the
        test seems to be too conservative. As a work-around, the binomial test
        is performed at a confidence level of 0.5 for the crit_dmodx (50%
        outliers).

        References
        ----------
        .. [1] L. Eriksson, E. Johansson, N. Kettaneh-Wold, J. Trygg, C.
        Wikström, and S. Wold. Multi- and Megavariate Data Analysis, Part I
        Basic Principles and Applications. Second Edition.
        """
        f_confidence = 0.50
        crit_dmodx = self.model.crit_dmodx(confidence=f_confidence)
        dmodx = self.model.dmodx(self.X_test)
        count = np.sum(dmodx > crit_dmodx)

        self._test_binom(1-f_confidence, self.n_tests, count)

    def test_crit_dhypx(self):
        """
        Test that number of outliers corresponds for crit_dhypx

        Performs a biomial statistical test. This test fails more often
        then the confidence level would predict. Potential explanation:
        crit_dhypx is not corrected for the bias in estimating the variance of
        the scores based on the training set [1]. This leads to a biased
        normalized dhypx calculations.

        References
        ----------
        .. [1] L. Eriksson, E. Johansson, N. Kettaneh-Wold, J. Trygg, C.
        Wikström, and S. Wold. Multi- and Megavariate Data Analysis, Part I
        Basic Principles and Applications. Second Edition.
        """
        f_confidence = 0.95
        crit_dhypx = self.model.crit_dhypx(confidence=f_confidence)
        dhypx = self.model.dhypx(self.X_test)
        count = np.sum(dhypx > crit_dhypx)
        self._test_binom(1-f_confidence, self.n_tests, count)

    def _test_binom(self, p, n_samples, n_positiv, false_positiv=0.005):
        """
        Perform a binomial test
        """
        pdist = binom(n_samples, p)
        limit_low = pdist.ppf(false_positiv/2)
        limit_high = pdist.ppf(1 - false_positiv/2)
        self.assertTrue(n_positiv >= limit_low)
        self.assertTrue(n_positiv <= limit_high)



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

"""
McrAR constraints

All classes need a transform method. Note, unlike sklearn, transform can copy
or overwrite input depending on copy attribute.
"""

from ._constraint import (
    Constraint,
    Nonneg,
    CumsumNonneg,
    ZeroEndPoints,
    ZeroCumSumEndPoints,
    Normalizer,
    CutBelow,
    CutAbove,
    CompressBelow,
    CompressAbove,
    ReplaceZeros,
    Unimodal,
)

__all__ = ['Constraint', 'Nonneg', 'CumsumNonneg', 'ZeroEndPoints',
           'ZeroCumSumEndPoints', 'Normalizer', 'CutBelow', 'CutAbove',
           'CompressBelow', 'CompressAbove', 'ReplaceZeros', 'Unimodal']



# Governed by two licenses
#
# Parts of the documentation:
# Adapted from sklearn, released under BSD-3 clause license
#
# Everything not adapted from sklearn:
# Copyright 2022 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

from .base import LVmixin
import numpy as np
import sklearn
from sklearn.decomposition import PCA as _PCA
from sklearn.pipeline import make_pipeline
from sklearn.model_selection import cross_val_score, KFold
import matplotlib.pyplot as plt


class PCA(_PCA, LVmixin):
    """
    Principal component analysis with added chemometric functionality

    Linear factorization of the data matrix X into scores and loadings
    (=components) similar to a truncated singular value decomposition.
    Next to the transformer capabilities, PCA provides additionally different
    metrics on the fitted latent variable model.

    """

    def __init__(
        self,
        n_components=2,
        *,
        copy=True,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        n_oversamples=10,
        power_iteration_normalizer="auto",
        random_state=None,
    ):
        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.n_oversamples = n_oversamples
        self.power_iteration_normalizer = power_iteration_normalizer
        self.random_state = random_state

    def fit(self, X, y=None):
        """
        Fit the model with X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        y : Ignored
            Ignored.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self.x_scores_ = super().fit_transform(X)
        self.x_residual_std_ = self._calculate_x_residual_std_(X)
        return self

    def fit_transform(self, X, y=None):
        """Fit the model with X and apply the dimensionality reduction on X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        y : Ignored
            Ignored.

        Returns
        -------
        X_new : ndarray of shape (n_samples, n_components)
            Transformed values.

        Notes
        -----
        This method returns a Fortran-ordered array. To convert it to a
        C-ordered array, use 'np.ascontiguousarray'.
        """

        self.x_scores_ = super().fit_transform(X)
        self.x_residual_std_ = self._calculate_x_residual_std_(X)
        return self.x_scores.copy()

    def score(self, X, y=None, scoring='r2'):
        """
        """
        X_hat = self.inverse_transform(self.transform(X))
        X_bar = np.mean(X, axis=0)
        predictor = np.sum((X-X_hat)**2)
        denominator = np.sum((X-X_bar)**2)
        score = 1-predictor/denominator

        return score

    @property
    def x_loadings_(self):
        """
        The loadings of `X` with shape (n_features, n_components).
        """
        return self.components_.T


def fit_pca(X, pipeline=None, cv_object=None, max_lv=10):
    r"""
    Auto-calibrate PCA model and generate analytical plots

    A PCA model is calibrated based on the maximization of the coefficient of
    determination during cross-validation (Q2). The function provides
    multiple plots for assessing the model quality. The first figure addresses
    the model performance during cross validation and the estimation of optimal
    number of latent variables by showing R2/Q2 values (R^2
    as bars, Q^2 as boxplots based on the individual rotations). The
    second figure shows four subplots with analytical information for the
    optimal model. The plotted figures are: a) observed versus predicted 2)
    predicted versus residuals 3) leverage versus residuals 4) Variable
    importance in projection (VIP) scores.

    Parameters
    ----------
    X : (n, m) ndarray
        Matrix of predictors. n samples x m predictors
    Y : (n, o) ndarray
        Matrix of responses. n samples x o responses
    pipeline : {None, sklearn.pipeline.Pipeline}
        A pipeline object providing a workflow of preprocessing and a
        PLSRegression model. The last entry must be a
        chemometrics.PCA instance.
    cv_object : {None, cv_object}
        An object providing guidance for cross-validation. Typically, it will
        be an instance of an sklearn.model_selection.BaseCrossValidator object.
    max_lv : int
        Number of latent variables up to which the cross-validation score will
        be screened.

    Returns
    -------
    pipeline : Pipeline
        The calibrated model pipeline
    summary : dict
        Summary of the model calibration.
    """
    if not pipeline:
        pipeline = make_pipeline(PCA())
    elif not (isinstance(pipeline, sklearn.pipeline.Pipeline)):
        raise TypeError(
            "pipeline argument provided is of type "
            + "{0} and not of type Pipeline.".format(type(pipeline))
        )
    elif not isinstance(pipeline[-1],
                        PCA):
        # check if pipeline ends with PCA
        raise TypeError(
            "Type of last object provided in pipline is "
            + "{0} but should be a PLSRegression".format(type(pipeline[-1]))
        )

    if not cv_object:
        cv_object = KFold(n_splits=5)

    # perform CV of model up to max_lv
    r2 = []
    q2 = []
    for n_lv in range(1, max_lv+1):
        pipeline[-1].n_components = n_lv
        q2.append(cross_val_score(pipeline, X, cv=cv_object))
        r2.append(pipeline.fit(X)[-1].explained_variance_ratio_.sum())

    q2 = np.stack(q2).T
    r2 = np.array(r2)

    # plot CV results
    fig_cv = plt.figure(figsize=(6, 6))
    plt.bar(np.arange(1, max_lv+1), r2, alpha=0.5)
    plt.boxplot(q2, positions=np.arange(1, max_lv+1))
    plt.xlabel('Latent variables')
    plt.ylabel('R2, Q2')

    # recover best q2 and adjust model accordingly
    pipeline[-1].n_components = np.argmax(np.median(q2, axis=0)) + 1
    pipeline = pipeline.fit(X)

    # plot PLS performance after preprocessing
    if len(pipeline) > 1:
        X_preprocessed = pipeline[:-1].transform(X)
    else:
        X_preprocessed = X
    pipeline[-1].distance_plot(X_preprocessed)
    fig_model = plt.gcf()

    # arrange return arguments
    analysis = {
        'r2': r2,
        'q2': q2,
        'q2_mean': np.mean(q2, axis=0),
        'q2_median': np.median(q2, axis=0),
        'optimal_lv': pipeline[-1].n_components,
        'figure_cv': fig_cv,
        'figure_model': fig_model
    }

    return pipeline, analysis



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
Testing mcr.constraints

"""

import numpy as np

from numpy.testing import assert_allclose

import chemometrics.mcr.constraint as constraint
import unittest


class TestConstraint(unittest.TestCase):
    def test_nonneg(self):
        A = np.array([[1, 2, 3], [-1, -2, -3], [1, 2, 3]])
        A_nn = np.array([[1, 2, 3], [0, 0, 0], [1, 2, 3]])

        constr_nn = constraint.Nonneg(copy=True)
        out = constr_nn.transform(A)
        assert_allclose(A_nn, out)

        constr_nn = constraint.Nonneg(copy=False)
        out = constr_nn.transform(A)
        assert_allclose(A_nn, A)

    def test_cumsumnonneg(self):
        """ Cum-Sum Nonnegativity Constraint """
        A = np.array([[2, -2, 3, -2], [-1, -2, -3, 7], [1, -2, -3, 7]])
        A_nn_ax1 = np.array([[2, 0, 3, -2], [0, 0, 0, 7], [1, 0, 0, 7]])
        A_nn_ax0 = np.array([[2, 0, 3, 0], [-1, 0, 0, 7], [1, 0, 0, 7]])

        # Axis -1
        constr_nn = constraint.CumsumNonneg(copy=True, axis=-1)
        out = constr_nn.transform(A)
        assert_allclose(A_nn_ax1, out)

        # Axis 0
        constr_nn = constraint.CumsumNonneg(copy=False, axis=0)
        out = constr_nn.transform(A)
        assert_allclose(A_nn_ax0, A)

    def test_zeroendpoints(self):
        """ 0-Endpoints Constraint """
        A = np.array([[1, 2, 3, 4],
                      [3, 6, 9, 12],
                      [4, 8, 12, 16]]).astype(float)
        A_ax1 = np.array([[0, 0, 0, 0],
                          [0, 0, 0, 0],
                          [0, 0, 0, 0]]).astype(float)
        A_ax0 = np.array([[0, 0, 0, 0],
                          [0.5, 1, 1.5, 2],
                          [0, 0, 0, 0]]).astype(float)

        # Axis 0
        constr_ax0 = constraint.ZeroEndPoints(copy=True, axis=0)
        out = constr_ax0.transform(A)
        assert_allclose(A_ax0, out)

        # Axis -1
        constr_ax1 = constraint.ZeroEndPoints(copy=True, axis=-1)
        out = constr_ax1.transform(A)
        assert_allclose(A_ax1, out)

        with self.assertRaises(TypeError):
            constr_ax1 = constraint.ZeroEndPoints(copy=True, axis=3)

        # Axis 0 -- NOT copies
        constr_ax0 = constraint.ZeroEndPoints(copy=False, axis=0)
        out = constr_ax0.transform(A)
        assert_allclose(A_ax0, A)

    def test_zeroendpoints_span(self):
        """ 0-Endpoints Constraint """
        A = np.array([[1, 2, 3, 4],
                      [3, 6, 9, 12],
                      [4, 8, 12, 16]]).astype(float)

        # Axis 1
        constr_ax1 = constraint.ZeroEndPoints(copy=True, axis=1, span=2)
        out = constr_ax1.transform(A)
        assert_allclose(out[:, [0, 1]].mean(axis=1), 0)
        assert_allclose(out[:, [1, 2]].mean(axis=1), 0)

        # Axis 0
        constr_ax0 = constraint.ZeroEndPoints(copy=True, axis=0, span=2)
        out = constr_ax0.transform(A)
        assert_allclose(out[[0, 1], :].mean(axis=0), 0)
        assert_allclose(out[[1, 2], :].mean(axis=0), 0)

        # effective an assert_not_equal
        assert_allclose([q != 0 for q in out[:, 0]], True)
        assert_allclose([q != 0 for q in out[:, -1]], True)

        # Axis 1 -- no copy
        constr_ax1 = constraint.ZeroEndPoints(copy=False, axis=1, span=2)
        out = constr_ax1.transform(A)
        assert_allclose(A[:, [0, 1]].mean(axis=1), 0)
        assert_allclose(A[:, [1, 2]].mean(axis=1), 0)

    def test_zerocumsumendpoints(self):
        """ Cum-Sum 0-Endpoints Constraint """
        A_diff1 = np.array([[1, 1, 1, 1],
                            [1, 1, 1, 1],
                            [1, 1, 1, 1]]).astype(float)
        A_diff0 = np.array([[3, 3, 3, 3],
                            [3, 3, 3, 3],
                            [3, 3, 3, 3]]).astype(float)

        # Axis 0
        constr_ax0 = constraint.ZeroCumSumEndPoints(copy=True, axis=0)
        out = constr_ax0.transform(A_diff0)
        assert_allclose(out, 0)
        assert_allclose(np.cumsum(out, axis=0), 0)

        # Axis -1
        constr_ax1 = constraint.ZeroCumSumEndPoints(copy=True, axis=-1)
        out = constr_ax1.transform(A_diff1)
        assert_allclose(out, 0)
        assert_allclose(np.cumsum(out, axis=1), 0)

        # Axis = -1 -- NOT copy
        constr_ax1 = constraint.ZeroCumSumEndPoints(copy=False, axis=-1)
        out = constr_ax1.transform(A_diff1)
        assert_allclose(A_diff1, 0)
        assert_allclose(np.cumsum(A_diff1, axis=1), 0)

    def test_zerocumsumendpoints_nodes(self):
        """ Cum-Sum 0-Endpoints Constraint with defined nodes"""

        A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(float)
        A_transform_ax0 = np.array([[0, 0, -1, 5],
                                    [0, 0, 0, -4],
                                    [0, 0, 1, -1]])
        A_transform_ax1 = np.array(
            [[-3, -2, -3, 8],
             [-1, 0, 0, 1],
             [-2, -1, 0, 3]]
            )

        # COPY Axis 0, node=0 (same as endpoints)
        constr_ax0 = constraint.ZeroCumSumEndPoints(copy=True,
                                                    nodes=[0],
                                                    axis=0)
        out = constr_ax0.transform(A)
        assert_allclose(out, A_transform_ax0)
        # assert_allclose(np.cumsum(out, axis=0), 0)

        # COPY Axis -1, node=0 (same as endpoints)
        constr_ax1 = constraint.ZeroCumSumEndPoints(copy=True,
                                                    nodes=[0],
                                                    axis=-1)
        out = constr_ax1.transform(A)
        assert_allclose(out, A_transform_ax1)
        # assert_allclose(np.cumsum(out, axis=1), 0)

        # OVERWRITE, Axis = 0, node=0 (same as endpoints)
        A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(float)
        constr_ax0 = constraint.ZeroCumSumEndPoints(copy=False,
                                                    nodes=[0],
                                                    axis=0)
        out = constr_ax0.transform(A)
        assert_allclose(A, A_transform_ax0)

        # OVERWRITE, Axis = 1, node=0 (same as endpoints)
        A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(float)
        constr_ax1 = constraint.ZeroCumSumEndPoints(copy=False,
                                                    nodes=[0],
                                                    axis=-1)
        out = constr_ax1.transform(A)
        assert_allclose(A, A_transform_ax1)

        # COPY, Axis = 0, Nodes = all
        A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(float)
        constr_ax0 = constraint.ZeroCumSumEndPoints(copy=True,
                                                    nodes=[0, 1, 2],
                                                    axis=0)
        out = constr_ax0.transform(A)
        assert_allclose(out, 0)

        # COPY, Axis = 1, Nodes = all
        A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(float)
        constr_ax1 = constraint.ZeroCumSumEndPoints(
            copy=True, nodes=[0, 1, 2, 3], axis=1
        )
        out = constr_ax1.transform(A)
        assert_allclose(out, 0)

        # OVERWRITE, Axis = 0, Nodes = all
        A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(float)
        constr_ax0 = constraint.ZeroCumSumEndPoints(copy=False,
                                                    nodes=[0, 1, 2],
                                                    axis=0)
        out = constr_ax0.transform(A)
        assert_allclose(A, 0)

        # OVERWRITE, Axis = 1, Nodes = all
        A = np.array([[1, 2, 1, 12], [1, 2, 2, 3], [1, 2, 3, 6]]).astype(float)
        constr_ax1 = constraint.ZeroCumSumEndPoints(
            copy=False, nodes=[0, 1, 2, 3], axis=1
        )
        out = constr_ax1.transform(A)
        assert_allclose(A, 0)

    def test_norm(self):
        """ Test normalization """
        # A must be dtype.float for in-place math (copy=False)
        constr_norm = constraint.Normalizer(axis=0, copy=False)

        # Axis must be 0,1, or -1
        with self.assertRaises(ValueError):
            constr_norm = constraint.Normalizer(axis=2, copy=False)

        A = np.array([[1, 2, 3], [-1, -2, -3], [1, 2, 3]], dtype=float)
        A_norm0 = A / A.sum(axis=0)[None, :]
        A_norm1 = A / A.sum(axis=1)[:, None]

        constr_norm = constraint.Normalizer(axis=0, copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_norm0, out)

        constr_norm = constraint.Normalizer(axis=1, copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_norm1, out)

        constr_norm = constraint.Normalizer(axis=-1, copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_norm1, out)

        constr_norm = constraint.Normalizer(axis=0, copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_norm0, A)

        A = np.array([[1, 2, 3], [-1, -2, -3], [1, 2, 3]], dtype=float)
        constr_norm = constraint.Normalizer(axis=1, copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_norm1, A)

    def test_norm_fixed_axes(self):
        # AXIS = 1
        A = np.array(
            [[0.0, 0.2, 1.0, 0.0],
             [0.25, 0.25, 0.0, 0.0],
             [0.3, 0.9, 0.6, 0.0]],
            dtype=float,
        )
        A_fix2_ax1 = np.array(
            [[0.0, 0.0, 1.0, 0.0], [0.5, 0.5, 0.0, 0.0], [0.1, 0.3, 0.6, 0.0]],
            dtype=float,
        )

        # Fixed axes must be integers
        with self.assertRaises(TypeError):
            constr_norm = constraint.Normalizer(axis=1, fix=2.2, copy=True)

        # Dtype must be integer related
        with self.assertRaises(TypeError):
            constr_norm = constraint.Normalizer(axis=1, fix=np.array([2.2]),
                                                copy=True)

        # COPY: True
        # Fix of type int
        constr_norm = constraint.Normalizer(axis=1, fix=2, copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, out)

        # Fix of type list
        constr_norm = constraint.Normalizer(axis=1, fix=[2, 3], copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, out)

        # Fix of type tuple
        constr_norm = constraint.Normalizer(axis=1, fix=(2), copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, out)

        # Fix of type ndarray
        constr_norm = constraint.Normalizer(axis=1, fix=np.array([2]),
                                            copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, out)

        # COPY: False
        A_fix2_ax1 = np.array(
            [[0.0, 0.0, 1.0, 0.0],
             [0.5, 0.5, 0.0, 0.0],
             [0.1, 0.3, 0.6, 0.0]],
            dtype=float,
        )

        # Fix of type int
        A = np.array(
            [[0.0, 0.2, 1.0, 0.0],
             [0.25, 0.25, 0.0, 0.0],
             [0.3, 0.9, 0.6, 0.0]],
            dtype=float,
        )
        constr_norm = constraint.Normalizer(axis=1, fix=2, copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, A)

        # Fix of type list
        A = np.array(
            [[0.0, 0.2, 1.0, 0.0],
             [0.25, 0.25, 0.0, 0.0],
             [0.3, 0.9, 0.6, 0.0]],
            dtype=float,
        )
        constr_norm = constraint.Normalizer(axis=1, fix=[2, 3], copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, A)

        # Fix of type tuple
        A = np.array(
            [[0.0, 0.2, 1.0, 0.0],
             [0.25, 0.25, 0.0, 0.0],
             [0.3, 0.9, 0.6, 0.0]],
            dtype=float,
        )
        constr_norm = constraint.Normalizer(axis=1, fix=(2), copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, A)

        # Fix of type ndarray
        A = np.array(
            [[0.0, 0.2, 1.0, 0.0],
             [0.25, 0.25, 0.0, 0.0],
             [0.3, 0.9, 0.6, 0.0]],
            dtype=float,
        )
        constr_norm = constraint.Normalizer(axis=1, fix=np.array([2]), copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, A)

        # AXIS = 0
        # Lazy, so just transposed
        A = np.array(
            [[0.0, 0.2, 1.0, 0.0],
             [0.25, 0.25, 0.0, 0.0],
             [0.3, 0.9, 0.6, 0.0]],
            dtype=float,
        ).T
        A_fix2_ax1 = np.array(
            [[0.0, 0.0, 1.0, 0.0],
             [0.5, 0.5, 0.0, 0.0],
             [0.1, 0.3, 0.6, 0.0]],
            dtype=float,
        ).T
        # COPY: True
        # Fix of type int
        constr_norm = constraint.Normalizer(axis=0, fix=2, copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, out)

        # Fix of type list
        constr_norm = constraint.Normalizer(axis=0, fix=[2, 3], copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, out)

        # Fix of type tuple
        constr_norm = constraint.Normalizer(axis=0, fix=(2), copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, out)

        # Fix of type ndarray
        constr_norm = constraint.Normalizer(axis=0, fix=np.array([2]),
                                            copy=True)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, out)

        # COPY: False
        A_fix2_ax1 = np.array(
            [[0.0, 0.0, 1.0], [0.5, 0.5, 0.0], [0.1, 0.3, 0.6]], dtype=float
        ).T

        # Fix of type int
        A = np.array(
            [[0.0, 0.2, 1.0], [0.25, 0.25, 0.0], [0.3, 0.9, 0.6]], dtype=float
        ).T
        constr_norm = constraint.Normalizer(axis=0, fix=2, copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, A)

        # Fix of type list
        A = np.array(
            [[0.0, 0.2, 1.0], [0.25, 0.25, 0.0], [0.3, 0.9, 0.6]], dtype=float
        ).T
        constr_norm = constraint.Normalizer(axis=0, fix=[2], copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, A)

        # Fix of type tuple
        A = np.array(
            [[0.0, 0.2, 1.0], [0.25, 0.25, 0.0], [0.3, 0.9, 0.6]], dtype=float
        ).T
        constr_norm = constraint.Normalizer(axis=0, fix=(2), copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, A)

        # Fix of type ndarray
        A = np.array(
            [[0.0, 0.2, 1.0], [0.25, 0.25, 0.0], [0.3, 0.9, 0.6]], dtype=float
        ).T
        constr_norm = constraint.Normalizer(axis=0, fix=np.array([2]),
                                            copy=False)
        out = constr_norm.transform(A)
        assert_allclose(A_fix2_ax1, A)

    def test_cut_below(self):
        """ Test cutting below (and not equal to) a value """
        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        A_transform = np.array([[0, 0, 0, 4],
                                [4, 5, 6, 7],
                                [7, 8, 9, 10]]).astype(float)

        constr = constraint.CutBelow(copy=True, value=4)
        out = constr.transform(A)
        assert_allclose(out, A_transform)

        # No Copy
        constr = constraint.CutBelow(copy=False, value=4)
        out = constr.transform(A)
        assert_allclose(A, A_transform)

    def test_cut_below_exclude(self):
        """ Test cutting below (and not equal to) a value """
        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        A_transform_excl_0_ax1 = np.array(
            [[1, 0, 0, 4], [4, 5, 6, 7], [7, 8, 9, 10]]
        ).astype(float)
        A_transform_excl_0_ax0 = np.array(
            [[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]
        ).astype(float)

        # COPY
        constr = constraint.CutBelow(copy=True, value=4, exclude=0,
                                     exclude_axis=1)
        out = constr.transform(A)
        assert_allclose(out, A_transform_excl_0_ax1)

        constr = constraint.CutBelow(copy=True, value=4, exclude=0,
                                     exclude_axis=0)
        out = constr.transform(A)
        assert_allclose(out, A_transform_excl_0_ax0)

        # OVERWRITE
        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        constr = constraint.CutBelow(copy=False, value=4, exclude=0,
                                     exclude_axis=1)
        out = constr.transform(A)
        assert_allclose(A, A_transform_excl_0_ax1)

        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        constr = constraint.CutBelow(copy=False, value=4, exclude=0,
                                     exclude_axis=0)
        out = constr.transform(A)
        assert_allclose(A, A_transform_excl_0_ax0)

    def test_cut_below_nonzerosum(self):
        """
        Test cutting below (and not equal to) a value with the constrain that
            no columns (axis=-1) will all become 0
        """
        A = np.array([[0.3, 0.7], [0.4, 0.6], [0.5, 0.5]])
        cutoff = 0.7
        A_correct = np.array([[0.0, 0.7], [0.4, 0.6], [0.5, 0.5]])

        constr = constraint.CutBelow(copy=True, value=cutoff, axis_sumnz=-1)
        out = constr.transform(A)
        assert_allclose(out, A_correct)

        constr = constraint.CutBelow(copy=False, value=cutoff, axis_sumnz=-1)
        constr.transform(A)
        assert_allclose(A, A_correct)

    def test_cut_below_nonzerosum_exclude(self):
        """
        Test cutting below (and not equal to) a value with the constrain that
            no columns (axis=-1) will all become 0
        """
        A = np.array([[0.3, 0.7], [0.3, 0.7], [0.7, 0.3]])
        cutoff = 0.7
        A_correct = np.array([[0.3, 0.7], [0.3, 0.7], [0.7, 0.0]])

        # COPY
        constr = constraint.CutBelow(
            copy=True, value=cutoff, axis_sumnz=-1, exclude=0, exclude_axis=-1
        )
        out = constr.transform(A)
        assert_allclose(out, A_correct)

        # OVERWRITE
        constr = constraint.CutBelow(
            copy=False, value=cutoff, axis_sumnz=-1, exclude=0, exclude_axis=-1
        )
        _ = constr.transform(A)
        assert_allclose(A, A_correct)

        # constr = constraint.CutBelow(copy=False, value=cutoff, axis_sumnz=-1)
        # constr.transform(A)
        # assert_allclose(A, A_correct)

    def test_cut_above_nonzerosum(self):
        """
        Test cutting above (and not equal to) a value with the constrain that
            no columns (axis=-1) will all become 0
        """
        A = np.array([[0.3, 0.7], [0.4, 0.6], [0.5, 0.5]])
        cutoff = 0.4
        A_correct = np.array([[0.3, 0.0], [0.4, 0.0], [0.5, 0.5]])

        constr = constraint.CutAbove(copy=True, value=cutoff, axis_sumnz=-1)
        out = constr.transform(A)
        assert_allclose(out, A_correct)

        # NO Copy
        A = np.array([[0.3, 0.7], [0.4, 0.6], [0.5, 0.5]])
        cutoff = 0.4
        A_correct = np.array([[0.3, 0.0], [0.4, 0.0], [0.5, 0.5]])

        constr = constraint.CutAbove(copy=False, value=cutoff, axis_sumnz=-1)
        constr.transform(A)
        assert_allclose(A, A_correct)

    def test_compress_below(self):
        """ Test compressing below (and not equal to) a value """
        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        A_transform = np.array([[4, 4, 4, 4],
                                [4, 5, 6, 7],
                                [7, 8, 9, 10]]).astype(float)

        constr = constraint.CompressBelow(copy=True, value=4)
        out = constr.transform(A)
        assert_allclose(out, A_transform)

        # No Copy
        constr = constraint.CompressBelow(copy=False, value=4)
        out = constr.transform(A)
        assert_allclose(A, A_transform)

    def test_cut_above(test_mcr_ideal_default):
        """ Test cutting above (and not equal to) a value """
        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        A_transform = np.array([[1, 2, 3, 4],
                                [4, 0, 0, 0],
                                [0, 0, 0, 0]]).astype(float)

        constr = constraint.CutAbove(copy=True, value=4)
        out = constr.transform(A)
        assert_allclose(out, A_transform)

        # No Copy
        constr = constraint.CutAbove(copy=False, value=4)
        out = constr.transform(A)
        assert_allclose(A, A_transform)

    def test_cut_above_exclude(self):
        """ Test cutting above (and not equal to) a value """
        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        A_transform_excl_0_ax1 = np.array(
            [[1, 2, 3, 4], [4, 0, 0, 0], [7, 0, 0, 0]]
        ).astype(float)
        A_transform_excl_2_ax0 = np.array(
            [[1, 2, 3, 4], [4, 0, 0, 0], [7, 8, 9, 10]]
        ).astype(float)

        constr = constraint.CutAbove(copy=True, value=4, exclude=0,
                                     exclude_axis=-1)
        out = constr.transform(A)
        assert_allclose(out, A_transform_excl_0_ax1)

        constr = constraint.CutAbove(copy=True, value=4, exclude=2,
                                     exclude_axis=0)
        out = constr.transform(A)
        assert_allclose(out, A_transform_excl_2_ax0)

        # No Copy
        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        constr = constraint.CutAbove(copy=False, value=4, exclude=0,
                                     exclude_axis=-1)
        _ = constr.transform(A)
        assert_allclose(A, A_transform_excl_0_ax1)

        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        constr = constraint.CutAbove(copy=False, value=4, exclude=2,
                                     exclude_axis=0)
        _ = constr.transform(A)
        assert_allclose(A, A_transform_excl_2_ax0)

    def test_compress_above(self):
        """ Test compressing above (and not equal to) a value """
        A = np.array([[1, 2, 3, 4], [4, 5, 6, 7], [7, 8, 9, 10]]).astype(float)
        A_transform = np.array(
            [[1, 2, 3, 4],
             [4, 4, 4, 4],
             [4, 4, 4, 4]]
            ).astype(float)

        constr = constraint.CompressAbove(copy=True, value=4)
        out = constr.transform(A)
        assert_allclose(out, A_transform)

        # No Copy
        constr = constraint.CompressAbove(copy=False, value=4)
        out = constr.transform(A)
        assert_allclose(A, A_transform)

    def test_replace_zeros(self):
        """ Test replace zeros using a single feature """
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        # Axis=0, feature = 0
        A_transform_ax0 = np.array(
            [
                [0.0, 0.0, 0.0, 1.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        # Axis=1, feature = 0
        A_transform_ax1 = np.array(
            [
                [1.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [1.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        # Axis 0, copy=True
        constr = constraint.ReplaceZeros(copy=True, axis=0, feature=0)
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax0)

        # Axis 1, copy=True
        constr = constraint.ReplaceZeros(copy=True, axis=1, feature=0)
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax1)

        # Axis 0, copy=FALSE
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        constr = constraint.ReplaceZeros(copy=False, axis=0, feature=0)
        out = constr.transform(A)
        assert_allclose(A, A_transform_ax0)

        # Axis 1, copy=FALSE
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        constr = constraint.ReplaceZeros(copy=False, axis=1, feature=0)
        out = constr.transform(A)
        assert_allclose(A, A_transform_ax1)

    def test_replace_zeros_multifeature(self):
        """ Replace zeros using multiple features """

        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        # Axis=0, feature = [0,1]
        A_transform_ax0 = np.array(
            [
                [0.0, 0.0, 0.0, 0.5],
                [0.25, 0.25, 0.0, 0.5],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        # Axis=1, feature = [0,1]
        A_transform_ax1 = np.array(
            [
                [0.5, 0.5, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.5, 0.5, 0.0, 0.0],
            ],
            dtype=float,
        )

        # Axis 0, copy=True, feature=[0,1]
        constr = constraint.ReplaceZeros(copy=True, axis=0, feature=[0, 1])
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax0)

        # Axis 1, copy=True, feature=[0,1]
        constr = constraint.ReplaceZeros(copy=True, axis=1, feature=[0, 1])
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax1)

        # Axis 1, copy=True, feature=np.array([0,1])
        constr = constraint.ReplaceZeros(copy=True, axis=1,
                                         feature=np.array([0, 1]))
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax1)

        # Axis 1, copy=True, feature=None
        constr = constraint.ReplaceZeros(copy=True, axis=1, feature=None)
        out = constr.transform(A)
        assert_allclose(out, A)

        # Axis 0, copy=FALSE, feature=[0,1]
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        constr = constraint.ReplaceZeros(copy=False, axis=0, feature=[0, 1])
        out = constr.transform(A)
        assert_allclose(A, A_transform_ax0)

        # Axis 1, copy=FALSE, feature=[0,1]
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        constr = constraint.ReplaceZeros(copy=False, axis=1, feature=[0, 1])
        out = constr.transform(A)
        assert_allclose(A, A_transform_ax1)

    def test_replace_zeros_non1fval(self):
        """ Test replace zeros using a single feature with fval != 1 """
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        # Axis=0, feature = 0
        A_transform_ax0 = np.array(
            [
                [0.0, 0.0, 0.0, 4.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        # Axis=1, feature = 0
        A_transform_ax1 = np.array(
            [
                [4.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [4.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        # Axis 0, copy=True
        constr = constraint.ReplaceZeros(copy=True, axis=0, feature=0,
                                         fval=4)
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax0)

        # Axis 1, copy=True
        constr = constraint.ReplaceZeros(copy=True, axis=1, feature=0, fval=4)
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax1)

        # Axis 0, copy=FALSE
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        constr = constraint.ReplaceZeros(copy=False, axis=0, feature=0, fval=4)
        out = constr.transform(A)
        assert_allclose(A, A_transform_ax0)

        # Axis 1, copy=FALSE
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        constr = constraint.ReplaceZeros(copy=False, axis=1, feature=0, fval=4)
        out = constr.transform(A)
        assert_allclose(A, A_transform_ax1)

    def test_replace_zeros_non1fval_multifeature(self):
        """ Test replace zeros using a 2 features with fval != 1 """
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        # Axis=0, feature = 0
        A_transform_ax0 = np.array(
            [
                [0.0, 0.0, 0.0, 2.0],
                [0.25, 0.25, 0.0, 2.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )
        # Axis=1, feature = 0
        A_transform_ax1 = np.array(
            [
                [2.0, 2.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [2.0, 2.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        # Axis 0, copy=True
        constr = constraint.ReplaceZeros(copy=True, axis=0, feature=[0, 1],
                                         fval=4)
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax0)

        # Axis 1, copy=True
        constr = constraint.ReplaceZeros(copy=True, axis=1, feature=[0, 1],
                                         fval=4)
        out = constr.transform(A)
        assert_allclose(out, A_transform_ax1)

        # Axis 0, copy=FALSE
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        constr = constraint.ReplaceZeros(copy=False, axis=0, feature=[0, 1],
                                         fval=4)
        out = constr.transform(A)
        assert_allclose(A, A_transform_ax0)

        # Axis 1, copy=FALSE
        A = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.25, 0.25, 0.0, 0.0],
                [0.3, 0.9, 0.6, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
            dtype=float,
        )

        constr = constraint.ReplaceZeros(copy=False, axis=1, feature=[0, 1],
                                         fval=4)
        out = constr.transform(A)
        assert_allclose(A, A_transform_ax1)

    def test_unimodal(self):
        n_samples = 500
        n_components = 3

        constr = constraint.Unimodal(copy=True)
        A = np.random.randn(n_samples, n_components)
        out = constr.transform(A)

        # assert that shape is not changed
        self.assertEqual(A.shape, out.shape)

        # assert that maximum remains in the same location
        argmaxA = np.argmax(A, axis=0)
        argmaxout = np.argmax(out, axis=0)
        assert_allclose(argmaxA, argmaxout)

        # assert that vectors are monotonically increasing befor max,
        # and monotonically decreasing after max.
        diff = np.diff(out, axis=0)
        argmaxout -= 1  # correct for shift due to diff
        mask_low = np.arange(diff.shape[0])[:, None] <= argmaxout[:, None].T
        mask_high = np.arange(diff.shape[0])[:, None] > argmaxout[:, None].T

        self.assertTrue(np.all(diff[mask_low] >= 0))
        self.assertTrue(np.all(diff[mask_high] <= 0))

        # check if columns can be excluded
        target_columns = [2]
        other_column = [0, 1]
        constr = constraint.Unimodal(columns=target_columns, copy=True)
        A = np.random.randn(n_samples, n_components)
        out = constr.transform(A)

        # assert that maximum remains in the same location
        argmaxA = np.argmax(A, axis=0)
        argmaxout = np.argmax(out, axis=0)
        assert_allclose(argmaxA, argmaxout)

        # assert that vectors are monotonically increasing befor max,
        # and monotonically decreasing after max.
        diff = np.diff(out, axis=0)
        argmaxout -= 1  # correct for shift due to diff
        mask_low = np.arange(diff.shape[0])[:, None] <= argmaxout[:, None].T
        mask_high = np.arange(diff.shape[0])[:, None] > argmaxout[:, None].T

        # remove untreated columns
        mask_low_target = mask_low.copy()
        mask_high_target = mask_high.copy()
        mask_low_target[:, other_column] = 0
        mask_high_target[:, other_column] = 0

        # test that treated columns are unimodal
        self.assertTrue(np.all(diff[mask_low_target] >= 0))
        self.assertTrue(np.all(diff[mask_high_target] <= 0))
        # test that some columns are not unimodal
        self.assertTrue(np.any(diff[mask_low] < 0))


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
Testing chemometrics.mcr.regressors

"""

import numpy as np

from chemometrics.mcr.metric import mse
import unittest


class TestMSE(unittest.TestCase):
    """
    Test mse calculation
    """

    def test_mse(self):
        A = np.ones((3, 3))
        B = np.eye(3)
        error = mse(None, None, A, B)
        self.assertEqual(error, 6/9)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.

import chemometrics as cm  # accounts for relativ path
import numpy as np
import unittest
import matplotlib


class TestPlot_colored_spectra(unittest.TestCase):
    r"""
    Test the `plot_colored_spectra` function.
    """

    def test_return_arguments(self):
        """
        Test if the correct shape and object type is returend.
        """
        n_series, n_variables = (10, 50)
        # generate dummy data
        x = np.arange(0, n_variables)
        Y = np.ones([n_series, n_variables]).T * np.linspace(0, 10,
                                                             num=n_series)
        lines = cm.plot_colored_series(Y, x=x)
        self.assertTrue(type(lines) == list)
        self.assertEqual(n_series, len(lines))
        for line in lines:
            self.assertTrue(type(line) == matplotlib.lines.Line2D)

    def tearDown(self):
        """
        Clean plots etc
        """
        matplotlib.pyplot.close()


class TestPlot_svd(unittest.TestCase):
    r"""
    Test the `plot_svd` function.
    """

    def test_return_arguments(self):
        n_series, n_variables = (50, 100)
        n_comp = 3
        n_eigenvalues = 12
        mean = np.zeros(n_variables)
        x = np.arange(n_variables)[:, None]
        dist = x.T - x
        cov = np.exp(-(dist / n_variables * 5)**2)

        # draw a few samples from gaussian process as data
        D = np.random.multivariate_normal(mean, cov, size=n_series)

        # execute function with artificial data
        cm.plot_svd(D, n_comp=n_comp, n_eigenvalues=n_eigenvalues)
        fig = matplotlib.pyplot.gcf()
        self.assertEqual(len(fig.axes), 3)
        self.assertEqual(len(fig.axes[0].lines), n_comp)
        self.assertEqual(len(fig.axes[1].lines), n_eigenvalues)
        self.assertEqual(len(fig.axes[2].lines), n_comp)

    def tearDown(self):
        """
        Clean plots etc
        """
        matplotlib.pyplot.close()



# Copyright 2021 Matthias Rüdt
#
# This file is part of chemometrics.
#
# chemometrics is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# chemometrics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with chemometrics.  If not, see <https://www.gnu.org/licenses/>.
#
# Additional information on the code history, original copyright holder and
# original license is available in ./license.md


"""
Test McrAR functionality
"""

import numpy as np

from numpy.testing import assert_equal

import unittest
from chemometrics.mcr import McrAR
import chemometrics.mcr.constraint as constraint
from chemometrics.mcr.regressor import OLS, NNLS
from sklearn.linear_model import Ridge


def dataset2():
    """ Setups dataset 2 components"""

    M = 21
    N = 21
    P = 101
    n_components = 2

    C_img = np.zeros((M, N, n_components))
    C_img[..., 0] = np.dot(np.ones((M, 1)), np.linspace(0, 1, N)[None, :])
    C_img[..., 1] = 1 - C_img[..., 0]

    St_known = np.zeros((n_components, P))
    St_known[0, 40:60] = 1
    St_known[1, 60:80] = 2

    C_known = C_img.reshape((-1, n_components))

    D_known = np.dot(C_known, St_known)

    return C_known, D_known, St_known


def dataset3():
    """ Setups dataset 3 components"""
    M = 21
    N = 21
    P = 101
    n_components = 3

    C_img = np.zeros((M, N, n_components))
    C_img[..., 0] = np.dot(np.ones((M, 1)), np.linspace(0, 1, N)[None, :])
    C_img[..., 1] = np.dot(np.linspace(0, 1, M)[:, None], np.ones((1, N)))
    C_img[..., 2] = 1 - C_img[..., 0] - C_img[..., 1]
    C_img = C_img / C_img.sum(axis=-1)[:, :, None]

    St_known = np.zeros((n_components, P))
    St_known[0, 30:50] = 1
    St_known[1, 50:70] = 2
    St_known[2, 70:90] = 3
    St_known += 1

    C_known = C_img.reshape((-1, n_components))

    D_known = np.dot(C_known, St_known)

    return C_known, D_known, St_known


class TestMcrAR(unittest.TestCase):
    """
    Test McrAR
    """

    def setUp(self):
        self.dataset2 = dataset2()
        self.dataset3 = dataset3()

    def test_mcr_ideal_default(self):
        """ Provides C/St_known so optimal should be 1 iteration """

        C_known, D_known, St_known = self.dataset2

        mcrar = McrAR()
        mcrar.fit(D_known, ST=St_known)
        self.assertLess(mcrar.n_iter_opt, 3)
        self.assertTrue(((mcrar.D_ - D_known)**2).mean() < 1e-10)
        self.assertTrue(((mcrar.D_opt_ - D_known)**2).mean() < 1e-10)

        mcrar.fit(D_known, C=C_known)
        self.assertLess(mcrar.n_iter_opt, 3)
        self.assertTrue(((mcrar.D_ - D_known)**2).mean() < 1e-10)
        self.assertTrue(((mcrar.D_opt_ - D_known)**2).mean() < 1e-10)

    def test_mcr_ideal_str_regressors(self):
        """ Test MCR with string-provded regressors"""

        C_known, D_known, St_known = self.dataset2

        mcrar = McrAR(c_regr='OLS', st_regr='OLS')
        mcrar.fit(D_known, ST=St_known, verbose=True)
        self.assertEqual(1, mcrar.n_iter_opt)
        assert isinstance(mcrar.c_regressor, OLS)
        assert isinstance(mcrar.st_regressor, OLS)

        mcrar = McrAR(c_regr='NNLS', st_regr='NNLS')
        mcrar.fit(D_known, ST=St_known)
        self.assertEqual(1, mcrar.n_iter_opt)
        assert isinstance(mcrar.c_regressor, NNLS)
        assert isinstance(mcrar.st_regressor, NNLS)
        assert ((mcrar.D_ - D_known)**2).mean() < 1e-10
        assert ((mcrar.D_opt_ - D_known)**2).mean() < 1e-10

        # Provided C_known this time
        mcrar = McrAR(c_regr='OLS', st_regr='OLS')
        mcrar.fit(D_known, C=C_known)

        # Turns out some systems get it in 1 iteration, some in 2
        # assert_equal(1, mcrar.n_iter_opt)
        self.assertEqual(True, mcrar.n_iter_opt <= 2)

        assert ((mcrar.D_ - D_known)**2).mean() < 1e-10
        assert ((mcrar.D_opt_ - D_known)**2).mean() < 1e-10

    def test_mcr_max_iterations(self):
        """ Test MCR exits at max_iter"""

        C_known, D_known, St_known = self.dataset2

        # Seeding with a constant of 0.1 for C, actually leads to a bad local
        # minimum; thus, the err_change gets really small with a relatively bad
        # error. The tol_err_change is set to None, so it makes it to max_iter.
        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=None, tol_n_increase=None,
                      tol_err_change=None, tol_n_above_min=None)
        mcrar.fit(D_known, C=C_known*0 + 0.1)
        assert mcrar.exit_max_iter_reached

    def test_mcr_tol_increase(self):
        """ Test MCR exits due error increasing above a tolerance fraction"""

        C_known, D_known, St_known = self.dataset2

        # Seeding with a constant of 0.1 for C, actually leads to a bad local
        # minimum; thus, the err_change gets really small with a relatively bad
        # error.
        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=0, tol_n_increase=None,
                      tol_err_change=None, tol_n_above_min=None)
        mcrar.fit(D_known, C=C_known*0 + 0.1)
        assert mcrar.exit_tol_increase

    def test_mcr_tol_n_increase(self):
        """
        Test MCR exits due iterating n times with an increase in error

        Note: On some CI systems, the minimum err bottoms out; thus,
        tol_n_above_min needed to be set to 0 to trigger a break.
        """

        C_known, D_known, St_known = self.dataset2

        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=None, tol_n_increase=0,
                      tol_err_change=None, tol_n_above_min=None)
        mcrar.fit(D_known, C=C_known*0 + 0.01)
        assert mcrar.exit_tol_n_increase

    def test_mcr_tol_err_change(self):
        """ Test MCR exits due error increasing by a value """

        C_known, D_known, St_known = self.dataset2

        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=None, tol_n_increase=None,
                      tol_err_change=1e-20, tol_n_above_min=None)
        mcrar.fit(D_known, C=C_known)
        assert mcrar.exit_tol_err_change

    def test_mcr_tol_n_above_min(self):
        """
        Test MCR exits due to half-terating n times with error above the
        minimum error.

        Note: On some CI systems, the minimum err bottoms out; thus,
        tol_n_above_min needed to be set to 0 to trigger a break.
        """

        C_known, D_known, St_known = self.dataset2

        mcrar = McrAR(max_iter=50, c_regr='OLS', st_regr='OLS',
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_increase=None, tol_n_increase=None,
                      tol_err_change=None, tol_n_above_min=0)
        mcrar.fit(D_known, C=C_known*0 + 0.1)
        assert mcrar.exit_tol_n_above_min

    def test_mcr_st_semilearned(self):
        """ Test when St items are fixed, i.e., enforced to be the same as the
        input, always """

        C_known, D_known, St_known = self.dataset3

        ST_guess = St_known.copy()
        ST_guess[2, :] = np.random.randn(St_known.shape[1])

        mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_err_change=1e-10)

        mcrar.fit(D_known, ST=ST_guess, st_fix=[0, 1])
        self.assertTrue(np.all(mcrar.ST_[0, :] == St_known[0, :]))
        self.assertTrue(np.all(mcrar.ST_[1, :] == St_known[1, :]))

    def test_mcr_c_semilearned(self):
        """ Test when C items are fixed, i.e., enforced to be the same as the
        input, always """

        C_known, D_known, St_known = self.dataset3

        # initialize concentration guess with true values
        C_guess = C_known.copy()
        # randomize third column
        C_guess[:, 2] = np.abs(np.random.randn(C_guess.shape[0])+0.1)

        mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_err_change=1e-10)

        mcrar.fit(D_known, C=C_guess, c_fix=[0, 1])
        # assert that column 1 and 2 of concentration remain unchanged
        assert_equal(mcrar.C_[:, 0], C_known[:, 0])
        assert_equal(mcrar.C_[:, 1], C_known[:, 1])

    def test_mcr_semilearned_both_c_st(self):
        """
        Test the special case when C & ST are provided, requiring C-fix ST-fix
        to be provided
        """
        C_known, D_known, St_known = self.dataset3

        C_guess = C_known.copy()
        # randomize third column
        C_guess[:, 2] = np.abs(np.random.randn(C_guess.shape[0])+0.1)

        mcrar = McrAR(max_iter=50, tol_increase=100, tol_n_increase=10,
                      st_constraints=[constraint.Nonneg()],
                      c_constraints=[constraint.Nonneg(),
                                     constraint.Normalizer()],
                      tol_err_change=1e-10)

        mcrar.fit(D_known, C=C_guess, ST=St_known,
                  c_fix=[0, 1], st_fix=[0], c_first=True)
        assert_equal(mcrar.C_[:, 0], C_known[:, 0])
        assert_equal(mcrar.C_[:, 1], C_known[:, 1])
        assert_equal(mcrar.ST_[0, :], St_known[0, :])

        # ST-solve first
        mcrar.fit(D_known, C=C_guess, ST=St_known, c_fix=[0, 1], st_fix=[0],
                  c_first=False)
        assert_equal(mcrar.C_[:, 0], C_known[:, 0])
        assert_equal(mcrar.C_[:, 1], C_known[:, 1])
        assert_equal(mcrar.ST_[0, :], St_known[0, :])

    def test_mcr_errors(self):

        # Providing both C and S^T estimates without C_fix and St_fix
        with self.assertRaises(TypeError):
            mcrar = McrAR()
            mcrar.fit(np.random.randn(10, 5), C=np.random.randn(10, 3),
                      ST=np.random.randn(3, 5))

        # Providing both C and S^T estimates without both C_fix and St_fix
        with self.assertRaises(TypeError):
            mcrar = McrAR()
            # Only c_fix
            mcrar.fit(np.random.randn(10, 5), C=np.random.randn(10, 3),
                      ST=np.random.randn(3, 5), c_fix=[0])

        with self.assertRaises(TypeError):
            mcrar = McrAR()
            # Only st_fix
            mcrar.fit(np.random.randn(10, 5), C=np.random.randn(10, 3),
                      ST=np.random.randn(3, 5), st_fix=[0])

        # Providing no estimates
        with self.assertRaises(TypeError):
            mcrar = McrAR()
            mcrar.fit(np.random.randn(10, 5))

        # Unknown regression method
        with self.assertRaises(ValueError):
            mcrar = McrAR(c_regr='NOTREAL')

        # regression object with no fit method
        with self.assertRaises(ValueError):
            mcrar = McrAR(c_regr=print)

    def test_props_features_samples_targets(self):
        """ Test mcrar properties for features, targets, samples """
        C_known, D_known, St_known = dataset2()

        mcrar = McrAR()
        mcrar.fit(D_known, ST=St_known)

        self.assertTrue(mcrar.n_targets == C_known.shape[-1])  # n_components
        self.assertTrue(mcrar.n_samples == D_known.shape[0])
        self.assertTrue(mcrar.n_features == D_known.shape[-1])

    def test_sklearn_linear_regression(self):
        """ Test MCR-AR works with sklearn-regressor """
        C_known, D_known, St_known = dataset2()

        mcrar = McrAR(c_regr=Ridge(), st_regr=Ridge())
        mcrar.fit(D_known, ST=St_known)

        # check that variables are set correctly w external regressor
        self.assertTrue(mcrar.n_targets == C_known.shape[-1])  # n_components
        self.assertTrue(mcrar.n_samples == D_known.shape[0])
        self.assertTrue(mcrar.n_features == D_known.shape[-1])

        # check that Ridge is used
        self.assertIsInstance(mcrar.c_regressor, Ridge)
        self.assertIsInstance(mcrar.st_regressor, Ridge)
        self.assertTrue(mcrar.c_regressor.n_features_in_ == C_known.shape[1])
        self.assertTrue(mcrar.st_regressor.n_features_in_ == C_known.shape[1])


if __name__ == '__main__':
    unittest.main()
